--DataSync: for ConstantGraph
-- (c) ConstantGraph 2021
-- www.constantgraph.com
local ssl = require("ssl")
local https = require("ssl.https")
local ltn12 = require("ltn12")
local lfs = require("lfs")
local json = require("ds-json")

local configData, dataMineConfig, manualConfig, loggedIds
local restricted = false
local MAX_RECORDS = 1000
local MANY_YEARS = 5*365*24*60*60
local LOGTIME_RAW = 7*24*60*60 -- create a new log file for every week
local dataDir = "/dataMine/"
local dataMineSync = dataDir.."dataSync.json"
local manualSync = "DataSyncConfig.json"
local apiKey = ""
local apiURL = "https://webhooks.mongodb-realm.com/api/client/v2.0/app/constantgraph-iwfeg/service/http/incoming_webhook/"
local dsDevice = ""
local DATASYNC_LOG_NAME = "DataSync: "
local SERVICE_ID = "urn:constantsphere-com:serviceId:DataSync"
local dsLuaVersion = "0.880"
local configUpdated = nil
local nextId = 1
local loginAttempts = 1
local DEFAULT_SERVICE_LIST = [[
	[{"Id":1,"Service":"urn:micasaverde-com:serviceId:DoorLock1","Variable":"Status"},
	{"Id":2,"Service":"urn:micasaverde-com:serviceId:EnergyMetering1","Variable":"CurrentTemperature"},
	{"Id":3,"Service":"urn:micasaverde-com:serviceId:EnergyMetering1","Variable":"KWH"},
	{"Id":4,"Service":"urn:micasaverde-com:serviceId:EnergyMetering1","Variable":"Watts"},
	{"Id":5,"Service":"urn:micasaverde-com:serviceId:HVAC_OperatingState1","Variable":"ModeState"},
	{"Id":6,"Service":"urn:micasaverde-com:serviceId:HaDevice1","Variable":"BatteryLevel"},
	{"Id":7,"Service":"urn:micasaverde-com:serviceId:HumiditySensor1","Variable":"CurrentLevel"},
	{"Id":8,"Service":"urn:micasaverde-com:serviceId:SecuritySensor1","Variable":"Tripped"},
	{"Id":9,"Service":"urn:upnp-org:serviceId:Dimming1","Variable":"LoadLevelStatus"},
	{"Id":10,"Service":"urn:upnp-org:serviceId:SwitchPower1","Variable":"Status"},
	{"Id":11,"Service":"urn:upnp-org:serviceId:TemperatureSensor1","Variable":"CurrentTemperature"}]
	]]
local serviceList = {}
local previousData = {}
local updatePeriod = 30

function Log(text,level)
	local logLevel = luup.variable_get(SERVICE_ID, "logLevel", dsDevice)
	logLevel = tonumber(logLevel) or 2
	level = level or 2 -- 1 = critical, 2 = info, 3 = debug
	if(level<=logLevel) then
		luup.log(DATASYNC_LOG_NAME .. "["..level.."]"..text)
	end
end

function _G.dump_debug(lul_request, lul_parameters)
	local html = ""
	html = html .. "-1-===========================================================================\n"
	local command = "cat /var/log/cmh/LuaUPnP.log | grep -i "..DATASYNC_LOG_NAME
	local f = io.popen(command, 'r')
	response = tostring(f:read('*a'))
	f:close(f)
	html = html .. response
	html = html .. "-2-===========================================================================\n"
	html = html .. "deviceId:" .. tostring(dsDevice) .. "\n"
	html = html .. "apiKey length:" .. string.len(luup.variable_get(SERVICE_ID, "apiKey", dsDevice) or "") .. "\n"
	html = html .. "dataMineDirectory:" .. (luup.variable_get(SERVICE_ID, "dataMineDirectory", dsDevice) or "@nil") .. "\n"
	html = html .. "syncWithDataMine:" .. (luup.variable_get(SERVICE_ID, "syncWithDataMine", dsDevice) or "@nil") .. "\n"
	html = html .. "Status:" .. (luup.variable_get(SERVICE_ID, "Status", dsDevice) or "@nil") .. "\n"
	html = html .. "loggedIds:" .. (luup.variable_get(SERVICE_ID, "loggedIds", dsDevice) or "@nil") .. "\n"
	html = html .. "serviceList:" .. (luup.variable_get(SERVICE_ID, "serviceList", dsDevice) or "@nil").. "\n"
	html = html .. "nextId:" .. (luup.variable_get(SERVICE_ID, "nextId", dsDevice) or "@nil").. "\n"
	html = html .. "updatePeriod:" .. (luup.variable_get(SERVICE_ID, "updatePeriod", dsDevice) or "@nil").. "\n"
	html = html .. "enabled:" .. (luup.variable_get(SERVICE_ID, "enabled", dsDevice) or "@nil").. "\n"
	html = html .. "logLevel:" .. (luup.variable_get(SERVICE_ID, "logLevel", dsDevice) or "@nil").. "\n"
	html = html .. "version:" .. (luup.variable_get(SERVICE_ID, "version", dsDevice) or "@nil").. "\n"
	html = html .. "-=============================================================================\n"
	return html, "text/plain"
end

function _G.restrictedMode(reason)
	restricted = true
	local isEnabled
	if(reason~=nil and reason ~="") then
		luup.variable_set(SERVICE_ID, "enabled", "0", dsDevice)
		luup.variable_set(SERVICE_ID, "Status", reason, dsDevice)
		Log("Going into restricted mode: "..reason)
	end
	isEnabled = luup.variable_get(SERVICE_ID, "enabled", dsDevice)
	if ((isEnabled == 1) or (isEnabled == "1")) then
		luup.variable_set(SERVICE_ID, "enabled", "1", dsDevice)
		Log("Retrying...")
		luup.variable_set(SERVICE_ID, "Status", "Retrying...", dsDevice)
		local aURL = luup.variable_get(SERVICE_ID, "apiURL", dsDevice)
		if(aURL ~= nil) then
			apiURL = tostring(aURL)
		end
		restricted = false
	else
		luup.call_delay('restrictedMode', 5)
	end
end

local function saveConfigFile()
	if(configData == nil) then
		configData = {}
		configData.channel = {}
	end
	local data = json.encode(configData) .. '\n'
	-- if execution gets aborted in the next few lines then the config file may get deleted
	local outf = io.open(dataMineSync, 'w')
	if(outf == nil) then
		Log("Unable to open config file for write :: " .. dataMineSync,1)
		restrictedMode("Unable to create sync file - check directory")
	else
		outf:write(data)
		outf:close()
	end
end

-- this is the dataSync config file that tracks which dataMine raw files have been uploaded
-- must be global as called by timer
function loadDataSyncConfig()
	Log("loading config file")
	local inf = io.open(dataMineSync, 'r')
	if(inf == nil) then
		-- likely file doesn't exist so create it
		Log("Unable to open config file for read :: " .. dataMineSync,2)
		saveConfigFile()
	else
		local line = inf:read("*all")
		inf:close()
		if(line == nil) then
			Log("Config file read failed :: " .. dataMineSync,1)
		elseif(string.len(line) < 4) then
			Log("Config file zero length:: " .. dataMineSync,1)
			saveConfigFile()
		else
			string.gsub(line, "%[%]", "%[{}%]")
			configData = json.decode(line)
			Log("decoded data sync config",2)
		end
	end
end

-- this is the dataSync config file that lists all of the channels to be monitored
-- in manual mode (when not using dataMine config)
local function loadManualConfig()
	loggedIds = {}
	serviceList = {}
	local dmLoggedIds = luup.variable_get(SERVICE_ID, "loggedIds", dsDevice)
	if(dmLoggedIds ~= nil) then
		Log("decoding manual data sync config",3)
		pcall(function() loggedIds = json.decode(dmLoggedIds) end)
	end

	local dmServiceList = luup.variable_get(SERVICE_ID, "serviceList", dsDevice)
	if(dmServiceList == nil) then
		dmServiceList = DEFAULT_SERVICE_LIST
		luup.variable_set(SERVICE_ID, "serviceList", dmServiceList, dsDevice)
	end
	Log("decoding manual service list",2)
	pcall(function() serviceList = json.decode(dmServiceList) end)
	nextId = luup.variable_get(SERVICE_ID, "nextId", dsDevice)
	if(nextId == nil) then 
		nextId = 1
		luup.variable_set(SERVICE_ID, "nextId", nextId, dsDevice)
	else 
		nextId = tonumber(nextId) 
	end
	updatePeriod = luup.variable_get(SERVICE_ID, "updatePeriod", dsDevice)
	if(updatePeriod ~= nil) then
		updatePeriod = tonumber(updatePeriod)
	end
end

local function findByDevice(device, service, variable)
	device=tostring(device)
	service=tostring(service)
	variable=tostring(variable)
	
	for k,v in pairs(dataMineConfig.Variables) do
		if((tostring(v.Device) == device) and (tostring(v.Service) == service) and (tostring(v.Variable)) == variable) then
		Log("Find by device d:"..device.." s:"..service.." v:"..variable.." Id:"..v.Id,3)
			return v
		end
	end
	Log("Find by device d:"..device.." s:"..service.." v:"..variable.." Id:not found",3)
	return ""
end

local function findById(id)
	id=tostring(id)
	for k,v in pairs(dataMineConfig.Variables) do
		if(tostring(v.Id) == id) then
			return v
		end
	end
	return ""
end

-- this is the dataMine config file that lists all the channels
function loadDataMineConfig()
	Log("checking dataMine config file attributes for changes",3)
	local attr,err = lfs.attributes(dataDir.."dataMineConfig.json")
	if(attr == nil) then
		Log("WARNING: Unable to check dataMine config file attributes :: " .. dataDir.."dataMineConfig.json",1)
		restrictedMode("Unable to load dataMine Config - check directory")
	elseif(attr.change ~= configUpdated) then
	
		apiKey = luup.variable_get(SERVICE_ID, "apiKey", dsDevice)
		configUpdated = attr.change
	
		Log("dataMine config file changed, reloading...",2)
		local inf = io.open(dataDir.."dataMineConfig.json", 'r')
		local line, response
		local path = apiURL.."config?apiKey="..apiKey
		local d
		
		if(inf == nil) then
			Log("WARNING: Unable to open dataMine config file for read :: " .. dataDir.."dataMineConfig.json",1)
			restrictedMode("Unable to load dataMine Config - check directory")
		else
			line = inf:read("*all")
			inf:close()
			if(line == nil or string.len(line) < 4) then
				Log("Config file zero length:: " .. dataDir.."dataMineConfig.json",1)
			else
				string.gsub(line, "%[%]", "%[{}%]")
				dataMineConfig = json.decode(line)
				Log("decoded dataMine config",2)
			end
			
			-- add ChannelIds to the config file
			for k,v in pairs(dataMineConfig.Graphs) do
				Log("config k:"..tostring(k).." v:"..tostring(v.Name),3)
				for k2,v2 in pairs(v.Channels) do
					d = findByDevice(v2.Device, v2.Service, v2.Variable )
					dataMineConfig.Graphs[k].Channels[k2].Id = d.Id
					if(dataMineConfig.Graphs[k].Channels[k2].Type == nil) then
						dataMineConfig.Graphs[k].Channels[k2].Type = d.Type
					end
				end
			end
			
			dataMineConfig.app = "vera"
			dataMineConfig.version = dsLuaVersion
			line = json.encode(dataMineConfig)
			
			local payload = string.gsub(line,"%%","%%25")
			Log("Path:"..path,3)	
			Log("Payload:"..payload,3)
			local command = "curl -H 'Content-Type: application/json' -s "..path.." -d '"..payload.."'"
			local f = io.popen(command, 'r')
			response = tostring(f:read('*a'))
			f:close(f)
			
			response = response:gsub( "\"", "")
			Log(response,2)			
		end
	end
end

function checkEnabled(wait)
	wait = tonumber(wait) or 5
	isEnabled = luup.variable_get(SERVICE_ID, "enabled", dsDevice)
	if(isEnabled ~= "1" and restricted == false) then -- catch any cases that we missed
		restrictedMode("Logging disabled")
		wait = 5
		Log("CheckEnabled back from restricted mode setting wait to "..wait,3)
	end
	if(restricted == true) then
		Log("calling checkEnabled with a wait of "..wait,3)
		luup.call_delay('checkEnabled', wait, "")
	else
		local dmSync = luup.variable_get(SERVICE_ID, "syncWithDataMine", dsDevice)
		if(dmSync == "1" or dmSync==1) then
			dataDir = luup.variable_get(SERVICE_ID, "dataMineDirectory", dsDevice)
			dataMineSync = dataDir.."dataSync.json"
			Log("calling transferFiles with a wait of "..wait,3)
			luup.call_delay('transferFiles', wait, "")
		else
			Log("calling transferManual with a wait of "..wait,3)
			luup.call_delay('transferManual', wait, "")
		end
	end
end

function transferManual()
	loadManualConfig()
	if(#loggedIds==0) then
		luup.variable_set(SERVICE_ID, "Status", "Select variables to start logging", dsDevice)
		return
	end
	apiKey = luup.variable_get(SERVICE_ID, "apiKey", dsDevice)
	local wait = updatePeriod
	local value,name,service,variable,id
	local payload='{"app":"vera","version":'..dsLuaVersion..',"channels":['
	local path = apiURL.."data?apiKey="..apiKey
	local line
	local count = 0
	local oldNextId = nextId
	Log("transfer values using manual process",3)
	for k,v in pairs(loggedIds) do
		service = serviceList[v.S].Service
		variable = serviceList[v.S].Variable
		value = luup.variable_get(service, variable, v.D)
		if pcall(function() name = luup.devices[v.D].description .. " - " .. variable end) then -- wrap in pcall as device may have been deleted
			id = v.I
			if(id>=nextId) then nextId = id+1 end
			if(id<0) then
				id = nextId
				v.I = id
				loggedIds[k].I = id
				nextId = nextId + 1
			end
			line = '{"id":'..id..',"v":"'..tostring(value)..'","Name":"'..name..'","Device":'..v.D..'},'
			Log(line,3)
			if(previousData[id] == line..apiKey) then
				Log("Data hasn't changed since last sent - excluding from payload",3)
			else
				Log("adding to payload",3)
				payload=payload..line
				previousData[id] = line..apiKey -- add APIKey so that variables are resent if API Key changes
				count = count + 1
			end
		else
			Log("Unable to find data for device:"..v.D.." it may have been deleted",2)
		end
	end
	if(oldNextId ~= nextId) then
		luup.variable_set(SERVICE_ID, "nextId", nextId, dsDevice)
		luup.variable_set(SERVICE_ID, "loggedIds", json.encode(loggedIds), dsDevice)
	end

	payload=payload:sub(1,-2)..']}'
	payload = payload:gsub( "'", "\'\\\'\'")
	if(count ==0) then 
		Log("no data has changed - no command sent",2)
	else
		local command = "curl -H 'Content-Type: application/json' -s "..path.." -d '"..payload.."'"
		Log("command:"..command,3)
		local f = io.popen(command, 'r')
		response = tostring(f:read('*a'))
		f:close(f)
		Log("response:"..tostring(response),2)

		if pcall(function() response = json.decode(response) end) then
			if(response ~= nil and response.error ~= nil) then 
				Log("Error returned from sending manual data:"..tostring(response.error),2)
				if(response.error_code == "RateLimit") then
					if(updatePeriod<600) then
						updatePeriod = updatePeriod + 60
						luup.variable_set(SERVICE_ID, "updatePeriod", updatePeriod, dsDevice)
					end
					wait = updatePeriod
					Log("Setting delay to "..updatePeriod.." seconds")
					mins = math.floor(wait / 60)
					if(mins <= 1) then
						luup.variable_set(SERVICE_ID, "Status", "Logging every "..tostring(wait).." seconds", dsDevice)
					else
						luup.variable_set(SERVICE_ID, "Status", "Logging every "..tostring(mins).." minutes", dsDevice)
					end
				elseif(response.error_code == "FunctionExecutionError") then
					-- if there is a server problem then wait a while
					Log("There is a problem accessing the service")
					luup.variable_set(SERVICE_ID, "Status", "There is a problem accessing the service", dsDevice)
					wait = 600
				elseif(response.error_code == "InvalidSession") then
					-- if there is an issue with logging  then wait a while
					if(loginAttempts == 2) then
						restrictedMode("Error logging in. Check API key.")
						loginAttempts = 1
						wait = 0
					else
						luup.variable_set(SERVICE_ID, "Status", "Error logging in. Retrying...", dsDevice)
						loginAttempts = loginAttempts + 1
					end
				end
			elseif(response == nil) then
				Log("response not decoded. Payload:"..tostring(payload),2)
				restrictedMode("Unable to form server request. Check logs.")
			else
				mins = math.floor(wait / 60)
				if(mins <= 1) then
					luup.variable_set(SERVICE_ID, "Status", "Logging every "..tostring(wait).." seconds", dsDevice)
				else
					luup.variable_set(SERVICE_ID, "Status", "Logging every "..tostring(mins).." minutes", dsDevice)
				end
			end
		end
	end
	checkEnabled(wait)
end

-- Run once at Luup engine startup.
function initialise(lul_device)

	-- Keep local copy of device number
	dsDevice = lul_device

	Log("Initialising dataSync System ("..dsLuaVersion..")")
	
	local isDisabled = luup.attr_get("disabled", dsDevice)
	if ((isDisabled == 1) or (isDisabled == "1")) then
		luup.variable_set(SERVICE_ID, "Status", "Plugin disabled", dsDevice)
		Log("Exiting initialisation as plug-in is disabled")
		return
	end

	luup.variable_set(SERVICE_ID, "version", dsLuaVersion, dsDevice)
	
	luup.variable_set(SERVICE_ID, "Status", "initialising", dsDevice)
	local aURL = luup.variable_get(SERVICE_ID, "apiURL", dsDevice)
	if(aURL == nil) then
      -- luup.variable_set(SERVICE_ID, "apiURL", apiURL, dsDevice)
	else
		apiURL = aURL
	end
	
	local aKey = luup.variable_get(SERVICE_ID, "apiKey", dsDevice)
	if(aKey == nil) then
		luup.variable_set(SERVICE_ID, "apiKey", apiKey, dsDevice)
	else
		apiKey = aKey
	end

	local dmSync = luup.variable_get(SERVICE_ID, "syncWithDataMine", dsDevice)
	if(dmSync == nil) then
		dmSync = 0
		luup.variable_set(SERVICE_ID, "syncWithDataMine", dmSync, dsDevice)
	end

	local dmEnabled = luup.variable_get(SERVICE_ID, "enabled", dsDevice)
	if(dmEnabled == nil) then
		dmEnabled = 0
		luup.variable_set(SERVICE_ID, "enabled", dmEnabled, dsDevice)
	end

	local logLevel = luup.variable_get(SERVICE_ID, "logLevel", dsDevice)
	if(logLevel == nil) then
		luup.variable_set(SERVICE_ID, "logLevel", "2", dsDevice)
	end

	local dmLoggedIds = luup.variable_get(SERVICE_ID, "loggedIds", dsDevice)
	if(dmLoggedIds == nil) then
		luup.variable_set(SERVICE_ID, "loggedIds", "[]", dsDevice)
		loggedIds = {}
	end

	local dmServiceList = luup.variable_get(SERVICE_ID, "serviceList", dsDevice)
	if(dmServiceList == nil) then
		dmServiceList = DEFAULT_SERVICE_LIST
		luup.variable_set(SERVICE_ID, "serviceList", dmServiceList, dsDevice)
	end
	pcall(function() serviceList = json.decode(dmServiceList) end)

	if(dmSync == "1" or dmSync==1) then
		local dmDir = luup.variable_get(SERVICE_ID, "dataMineDirectory", dsDevice)
		if(dmDir == nil) then
			luup.variable_set(SERVICE_ID, "dataMineDirectory", dataDir, dsDevice)
		else
			dataDir = dmDir
			dataMineSync = dataDir.."dataSync.json"
		end
	else
		luup.variable_set(SERVICE_ID, "updatePeriod", updatePeriod, dsDevice)
	end

	luup.register_handler("dump_debug", "dsDebug_"..tostring(dsDevice))
	checkEnabled(5)
end

local function sendChannel(channelId,channelName)
	local response ="NODATA"
	if(channelId == nil) then channelId = 0 end
	channelId=tostring(channelId)
	if(configData.channel[channelId] == nil) then 
		Log("creating new record for "..channelId)
		configData.channel[channelId] = {} 
	end

	local Stop = os.time() -- dataMineConfig.LastWrite
	local Start = configData.channel[channelId].timestamp
	local Variable = findById(channelId)
	local alpha = (Variable.Alpha == 1)
	apiKey = luup.variable_get(SERVICE_ID, "apiKey", dsDevice)

	Log("Start is:"..tostring(Start).." Stop is:"..tostring(Stop).." channel alpha is:"..tostring(alpha),3)

	if(Stop == nil) then Stop =	os.time() end
	if(Start == nil) then Start = Stop - MANY_YEARS end  -- if no start time then go back a few years
	
	if(Start>=Stop) then return response,0 end

	local WeekNum  = math.floor(Start / LOGTIME_RAW)
	local WeekEnd  = math.floor(Stop  / LOGTIME_RAW)
	
	local filename 
	local recordCount = 0

	local l_time = Start
	local path = apiURL.."timedata?apiKey="..apiKey

	local payload='{"app":"vera","version":'..dsLuaVersion..',"channels":[{"id":'..channelId..',"values":['
	repeat
		filename = dataDir.."database/"..channelId.."/raw/"..WeekNum..".txt"
		local inf = io.open(filename, 'r')
		if(inf ~= nil) then
			while true do
				line = inf:read("*line")
				if(line == nil or recordCount>=MAX_RECORDS) then
					inf:close()
					break
				end
				if(alpha) then 
					l_time,l_val = 0+line:sub(1,10),line:sub(12)
					if(Variable.Lookup~=nil and l_time > Start) then
						l_val = Variable.Lookup[l_val]
						if(l_val == nil) then l_val  = line:sub(12) end
						payload=payload..'{"t":'..l_time..',"v":'..l_val..'},'
						recordCount = recordCount + 1
					end
				elseif pcall(function() l_time,l_val = 0+line:sub(1,10),0+line:sub(12) end) and l_time > Start then 
					payload=payload..'{"t":'..l_time..',"v":'..l_val..'},'
					recordCount = recordCount + 1
				end
			end
		end

		if(recordCount>=MAX_RECORDS) then break end
		WeekNum = WeekNum + 1
		
	until WeekNum>WeekEnd
	
	payload=payload:sub(1,-2)..']}]}'
	
	if(recordCount>0) then
		Log("found "..recordCount.." records for channel "..channelId.." ("..channelName..") between "..os.date("%c",Start).." and "..os.date("%c",l_time))
		Log("payload:"..payload,3)
		
		local command = "curl -H 'Content-Type: application/json' -s "..path.." -d '"..payload.."'"
		local f = io.popen(command, 'r')
		response = tostring(f:read('*a'))
		f:close(f)

		Log("Returned response:"..response,3)
		Log("Path:"..path,3)
		if pcall(function() response = json.decode(response) end) then
			if(response.error == nil) then 
				configData.channel[channelId].timestamp = l_time
			elseif(response.error_code == "FunctionExecutionError") then
				-- if there is a server problem then wait a while
				Log("There is a problem accessing the service")
			elseif(response.error_code == "InvalidSession") then
				-- if there is an issue with logging in then wait a while
				if(loginAttempts == 2) then
					restrictedMode("Error logging in. Check API key.")
					loginAttempts = 1
					wait = 0
				else
					luup.variable_set(SERVICE_ID, "Status", "Error logging in. Retrying...", dsDevice)
					loginAttempts = loginAttempts + 1
				end
			elseif(response.error_code == "AccessDenied") then
				-- if there is an issue with the subscription then wait a while
				restrictedMode("Error accessing service - check subscription level")
			else
				configData.channel[channelId].error = response.error
				Log("Error returned from sending data to channel:"..channelId.." error:"..response.error,2)
			end
		end
	else
		configData.channel[channelId].timestamp = Stop
	end
	return response, recordCount
end

function transferFiles()
	loadDataMineConfig()
	local start_time = os.time()
	local channelcount=0
	local recordcount=0
	local wait = 60
	local r=0
	local response
	if(dataMineConfig~=nil) then
		loadDataSyncConfig()
		luup.variable_set(SERVICE_ID, "Status", "Synchronising with dataMine 2", dsDevice)
		for k,v in pairs(dataMineConfig.Variables) do 
			Log("transfer files channel:"..v.Id,3)
			response, r = sendChannel(v.Id,v.Name)
			recordcount=recordcount+r
			if(r == MAX_RECORDS) then
				--wait = 0 -- don't wait for next pass if stopped for max records
				Log("setting wait time to "..wait.." as max records sent",2)
			end
			if(response == nil or response.error_code == "FunctionExecutionError") then
				Log("no or bad response when sending data for channel:"..v.Id.." - waiting 5 minutes")
				luup.variable_set(SERVICE_ID, "Status", "Issue synchronising data - temporarily paused", dsDevice)
				wait = 300
				break
			elseif(response.error ~=nil) then
				Log("error when sending data for channel:"..v.Id,3)
			elseif(response.inserted ~=nil) then -- success
				channelcount=channelcount+1
			end
		end

		end_time = os.time()
		Log("Completed pass of Sync Operation in "..os.difftime(end_time-start_time).." seconds. "..recordcount.." records across "..channelcount.." channels transferred")
		saveConfigFile()
	end
	checkEnabled(wait)
end